part of model_notifier;

/// A hook that listens to a provider and returns its current value.
///
/// As a usage example, consider:
///
/// ```dart
/// final helloWorldProvider = Provider((_) => 'Hello world');
/// ```
///
/// We can then use [useProvider] to listen to `helloWorldProvider` like so:
///
/// ```dart
/// class Example extends HookWidget {
///   @override
///   Widget build(BuildContext context) {
///     final value = useProvider(helloWorldProvider);
///     return Text(value); // Hello world
///   }
/// }
/// ```
///
/// See also:
///
/// - [Provider]/`select`, for filtering unwanted rebuilds.
const useProvider = _UseProviderBuilder();

class _UseProviderBuilder {
  const _UseProviderBuilder();

  /// Reads a provider without listening to it.
  ///
  /// **AVOID** calling [provider] inside build if the value is used only for events:
  ///
  /// ```dart
  /// Widget build(BuildContext context) {
  ///   // counter is used only for the onPressed of RaisedButton
  ///   final counter = useProvider.read(counterProvider);
  ///
  ///   return RaisedButton(
  ///     onPressed: () => counter.increment(),
  ///   );
  /// }
  /// ```
  ///
  /// While this code is not bugged in itself, this is an anti-pattern.
  /// It could easily lead to bugs in the future after refactoring the widget
  /// to use `counter` for other things, but forget to change [provider] into [Consumer]/`useProvider.watch`.
  T read<T>(ProviderBase<T> provider) {
    final container = ProviderScope.containerOf(useContext(), listen: false);
    return container.read(provider);
  }

  /// A hook that listens to a provider and returns its current value.
  ///
  /// As a usage example, consider:
  ///
  /// ```dart
  /// final helloWorldProvider = Provider((_) => 'Hello world');
  /// ```
  ///
  /// We can then use [useProvider] to listen to `helloWorldProvider` like so:
  ///
  /// ```dart
  /// class Example extends HookWidget {
  ///   @override
  ///   Widget build(BuildContext context) {
  ///     final value = useProvider(helloWorldProvider);
  ///     return Text(value); // Hello world
  ///   }
  /// }
  /// ```
  ///
  /// See also:
  ///
  /// - [Provider]/`select`, for filtering unwanted rebuilds.
  T watch<T>(ProviderListenable<T> provider) => call(provider);

  T call<T>(ProviderListenable<T> provider) {
    final container = ProviderScope.containerOf(useContext());
    return use(_ProviderHook<T>(container, provider));
  }
}

class _ProviderHook<T> extends Hook<T> {
  const _ProviderHook(this._container, this._providerListenable);

  final ProviderContainer _container;
  final ProviderListenable<T> _providerListenable;

  @override
  _ProviderHookState<T> createState() => _ProviderHookState();
}

class _ProviderHookState<T> extends HookState<T, _ProviderHook<T>> {
  ProviderSubscription<T>? _link;

  @override
  void initHook() {
    super.initHook();
    _listen();
  }

  void _listen() {
    _link?.close();
    _link =
        hook._container.listen<T>(hook._providerListenable, _handledOnChanged);
  }

  void _handledOnChanged(T state) {
    markMayNeedRebuild();
  }

  @override
  T build(BuildContext context) {
    return _link!.read();
  }

  @override
  void didUpdateHook(_ProviderHook<T> oldHook) {
    super.didUpdateHook(oldHook);
    assert(
      oldHook._providerListenable.runtimeType ==
          hook._providerListenable.runtimeType,
      'The provider listened cannot change',
    );

    if (oldHook._container != hook._container) {
      _listen();
    } else if (oldHook._providerListenable != hook._providerListenable) {
      _listen();
    }
  }

  @override
  void dispose() {
    _link!.close();
    super.dispose();
  }
}
